/* $Id: e2_keybinding.c 1750 2006-03-11 07:19:05Z tpgww $
 *
 * Copyright (C) 2004-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelfm2.
 * emelfm2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelfm2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_keybinding.c 
 * @brief functions for handling key bindings
 **/

#include "e2_keybinding.h"
#include "e2_option_tree.h"
#include <string.h>

E2_OptionSet *opt_keys;
//keys is a list of E2_KeybindingRuntime structs, each for a 'category' of bindings
GList *keys = NULL;
#ifdef E2_IDLE_KEYBINDING
//keys is a pointer for use with the keys GList
GList *keys_cur = NULL;
#endif

  /*****************/
 /***** utils *****/
/*****************/

/**
 * @brief cleanup bound-keys data in list @a keys
 * Recurses to clean chained keys, if any
 * @param keys pointer to a bound-keys list
 * @return
 **/
static void _e2_keybinding_free_keyslist (GList *usedkeys)
{
	GList *list;
	for (list = usedkeys; list != NULL; list = list->next)
	{
		E2_KeyRuntime *k = list->data;
		if (k != NULL)
		{
			if (k->chained != NULL)
				_e2_keybinding_free_keyslist (k->chained);
			g_free (k->action);
			g_free (k->action_data);	//CHECKME can this be NULL ?
			g_free (k);
		}
	}
	g_list_free (usedkeys);
}
/**
 * @brief cleanup data for the bound keys for the category described by @a rt
 * @param rt pointer to keybinding category data struct
 * @return
 **/
static void _e2_keybinding_free_keys (E2_KeybindingRuntime *rt)
{
	if (rt->keys != NULL)
	{
		_e2_keybinding_free_keyslist (rt->keys);
		rt->keys = NULL;
	}
}
/**
 * @brief translate keybinding model row into a listed struct
 * @param model model for the bindings treestore
 * @param iter pointer to iter to be used to interrogage @a model
 * @param keys pointer to list to which the created struct will be appended
 * @return
 **/
static void _e2_keybinding_sync_one (GtkTreeModel *model, GtkTreeIter *iter, GList **keys)
{
	gchar *cat, *key, *action, *action_data;
	gboolean cont;
	gtk_tree_model_get (model, iter, 0, &cat, 1, &key, 2, &cont, 3 , &action,
		4, &action_data, -1);
	if (((cat != NULL) && (cat[0] != '\0')) || (key == NULL))
	{
		g_free (cat);
		g_free (key);
		g_free (action);
		g_free (action_data);
		return;
	}
	GdkModifierType mask;
	guint keyval;
	gtk_accelerator_parse (key, &keyval, &mask);
	if ((keyval != 0) || (mask != 0))
	{
		/*for letters, when a <Shift> is pressed, or after <Lock>, 
		gtk actually returns the upper-case code, = lower-case - 0x20
		if <Shift> , there will also be GDK_SHIFT_MASK
		but e2's config widget always provides the <Shift>lower-case pair
		we need to work arouind this mismatch */
//		printd (DEBUG, "key - %s - %d - %d", key, keyval, mask);
		if (keyval >= GDK_a && keyval <= GDK_z && mask == GDK_SHIFT_MASK)
			keyval -= 0x20;
		E2_KeyRuntime *k = g_malloc (sizeof (E2_KeyRuntime));
		k->keyval = keyval;
		k->mask = mask;
		k->action = action;
		k->action_data = action_data;
		k->cont = cont;
		k->chained = NULL;
		GtkTreePath *path = gtk_tree_model_get_path (model, iter);
		gtk_tree_path_down (path);
		GtkTreeIter child_iter;
		while (gtk_tree_model_get_iter (model, &child_iter, path))
		{
			_e2_keybinding_sync_one (model, &child_iter, &k->chained);
			gtk_tree_path_next (path);
		}
		gtk_tree_path_free (path);
		*keys = g_list_append (*keys, k);
	} else
	{
		g_free (action);
		g_free (action_data);
	}
	g_free (cat);
	g_free (key);
	return;
}
/**
 * @brief translate, from options tree to glist, the keybindings data for a bindings category
 * @param rt pointer to data struct for the keybinding category to be processed
 * @return TRUE if the process completed properly
 **/
static
#ifdef E2_IDLE_KEYBINDING
gboolean
#else
void
#endif
 _e2_keybinding_sync (E2_KeybindingRuntime *rt)
{
	if (!gtk_tree_row_reference_valid (rt->ref))
	{
		printd (WARN, "internal keybinding error, row reference not valid anymore");
#ifdef E2_IDLE_KEYBINDING
		return FALSE;
#else
		return;
#endif
	}
	GtkTreePath *path = gtk_tree_row_reference_get_path (rt->ref);
	if (path == NULL)
	{
		printd (WARN, "internal keybinding error, path is NULL");
#ifdef E2_IDLE_KEYBINDING
		return FALSE;
#else
		return;
#endif
	}

	_e2_keybinding_free_keys (rt);

	gtk_tree_path_down (path);
	GtkTreeIter iter;
	while (gtk_tree_model_get_iter (rt->model, &iter, path))
	{
		_e2_keybinding_sync_one (rt->model, &iter, &rt->keys);
		gtk_tree_path_next (path);
	}
	gtk_tree_path_free (path);
#ifdef E2_IDLE_KEYBINDING
	return TRUE;
#else
	return;
#endif
}

  /*********************/
 /***** callbacks *****/
/*********************/

/**
 * @brief (de)sensitize option tree buttons for selected option tree row
 * Config dialog page buttons are de-sensitized if the row is a category.
 * @param selection pointer to selection
 * @param model
 * @param path
 * @param path_currently_selected
 * @param set data struct for the keybings option
 * @return TRUE always (the row is always selectable)
 **/
static gboolean _e2_keybinding_tree_selection_check_cb
	(GtkTreeSelection *selection, GtkTreeModel *model, GtkTreePath *path,
	gboolean path_currently_selected, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (set->ex.tree.model, &iter, path))
	{
		gchar *cat;
		gtk_tree_model_get (set->ex.tree.model, &iter, 0, &cat, -1);
		gboolean retval = (cat[0] == '\0');
		g_free (cat);
		GtkTreeView *view = gtk_tree_selection_get_tree_view (selection);
		e2_option_tree_adjust_buttons (view, retval);
	}
	return TRUE;
}
/**
 * @brief ensure that cells in the same row as any category name are hidden
 *
 * Checks whether @a iter in @a model has an empty string in column 0.
 * If not (i.e. it is a category row), the cell content is hidden,
 *
 * @param model treemodel for the keybindings option tree
 * @param iter pointer to iter with data for the current model row
 * @param cell cell renderer UNUSED
 * @param data pointer to data UNUSED
 * @return TRUE (cell is visible) if row is not a category name
 **/
static gboolean _e2_keybinding_visible_check_cb (GtkTreeModel *model,
	GtkTreeIter *iter, GtkCellRenderer *cell, gpointer data)
{
	gchar *cat;
	gtk_tree_model_get (model, iter, 0, &cat, -1);
	gboolean retval = (cat[0] == '\0');
	g_free (cat);
	return retval;
}
/**
 * @brief decide whether tree row is draggable
 * Checks whether column 0 of the current row has a null string
 * ie the row is not a category. If so, the row is draggable
 * @param drag_source GtkTreeDragSource data struct
 * @param path tree path to a row on which user is initiating a drag
 * @return TRUE if the row can be dragged
 **/
static gboolean _e2_keybinding_tree_draggable_check_cb
	(GtkTreeDragSource *drag_source, GtkTreePath *path)
{
	if (!GTK_IS_TREE_MODEL (drag_source))
		return TRUE;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_source);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (model, &iter, path))
	{
		gchar *cat;
		gtk_tree_model_get (model, &iter, 0, &cat, -1);
		gboolean retval = (cat[0] == '\0');
		g_free (cat);
		return retval;
	}
	return TRUE;
}
/**
 * @brief cancel the chained keys specified in category binding detailed by @a rt
 * This func is mainloop idle callbacK
 * @param rt pointer to keybinding data struct
 * @return FALSE always
 **/
static gboolean _e2_keybinding_remove_chained_idle_cb (E2_KeybindingRuntime *rt)
{
	rt->chained = NULL;
	rt->timeout_id = 0;
	return FALSE;
}
/**
 * @brief find and run the action that's bound to a key
 * @param widget the widget to which the binding was bound
 * @param event data struct for the event
 * @param rt pointer to keybinding data struct
 * @return TRUE if the continuation flag of the bound key is TRUE
 **/
static gboolean _e2_keybinding_key_press_cb (GtkWidget *widget,
	GdkEventKey *event, E2_KeybindingRuntime *rt)
{
	guint modifiers = gtk_accelerator_get_default_mod_mask ();
	modifiers = event->state & modifiers;
	printd (DEBUG, "key 1: %d - %d", event->keyval, modifiers);
	/*workaround for matching our storage of <Shift>-letter
	with gtk's approach to returning keycodes
	See comments in _e2_keybinding_sync_one () */
	if (event->state & GDK_LOCK_MASK)
		modifiers |= GDK_SHIFT_MASK;
	GList *list;
	for (list = rt->chained; list != NULL; list = g_list_next (list))
	{
		E2_KeyRuntime *k = list->data;
		if ((k->keyval == event->keyval) && (k->mask == modifiers))
		{
			if (k->chained != NULL)
			{
				rt->chained = k->chained;
				if (rt->timeout_id != 0)
					g_source_remove (rt->timeout_id);
				rt->timeout_id = g_timeout_add (e2_option_int_get ("keybindings-timeout"),
					(GSourceFunc) _e2_keybinding_remove_chained_idle_cb, rt);
			} else
			{
				if (rt->timeout_id != 0)
					g_source_remove (rt->timeout_id);
				_e2_keybinding_remove_chained_idle_cb (rt);
			}
				
			e2_action_run_simple_from (k->action, k->action_data, widget);
			if (!k->cont)
				return TRUE;
		}
	}
	for (list = rt->keys; list != NULL; list = g_list_next (list))
	{
		E2_KeyRuntime *k = list->data;
		if ((k->keyval == event->keyval) && (k->mask == modifiers))
		{
			if (k->chained != NULL)
			{
				rt->chained = k->chained;
				if (rt->timeout_id != 0)
					g_source_remove (rt->timeout_id);
				rt->timeout_id = g_timeout_add (e2_option_int_get ("keybindings-timeout"),
					(GSourceFunc) _e2_keybinding_remove_chained_idle_cb, rt);
			}
			e2_action_run_simple_from (k->action, k->action_data, widget);
			if (!k->cont)
				return TRUE;
		}
	}
	return FALSE;
}
#ifdef E2_IDLE_KEYBINDING
// CHECKME don't do idle-sync's now
/**
 * @brief convert bindings data for a bindings category
 * This func is mainloop idle callback - WHY SYNC THAT WAY?
 * @param data UNUSED pointer to data specified when idle was added
 * @return TRUE (so the fn continues to be called) if ??
 **/
static gboolean _e2_keybinding_idle_sync_cb (gpointer data)
{
	if (keys_cur == NULL)
		keys_cur = keys;
	if (keys_cur != NULL)
	{
//		gdk_threads_enter ();	//protect idle fn against thread glitches
		//sync a single KeybindingRuntime
		gint retval = _e2_keybinding_sync ((E2_KeybindingRuntime *) keys_cur->data);
//		gdk_threads_leave ();
		keys_cur = g_list_next (keys_cur);
		if (keys_cur != NULL)
			return retval;
	}
	return FALSE;	//remove this fn from event sources 
}
/**
 * @brief setup for converting keybinding data, after the bindindgs treestore is altered
 * @param model UNUSED
 * @param arg1  UNUSED
 * @param arg2  UNUSED
 * @param data  UNUSED
 * @return
 **/
static void _e2_keybinding_model_changed_cb
		(GtkTreeModel *model,
		GtkTreePath *arg1,
		GtkTreeIter *arg2,
		gpointer data)
{
	keys_cur = NULL;
	g_idle_add ((GSourceFunc) _e2_keybinding_idle_sync_cb, NULL);
}
#endif  //def E2_IDLE_KEYBINDING

  /*******************/
 /***** public ******/
/******************/

/**
 * @brief check keybindings list to see if @a name is registered
 * @param name the name of the binding category to search for
 * @return glist entry that matches @a name, or NULL if no match
 **/
GList *e2_keybinding_find_by_name (const gchar *name)
{
	GList *list;
	for (list=keys; list!=NULL; list=list->next)
	{
		if (g_str_equal (name, ((E2_KeybindingRuntime *) list->data)->name))
			break;
	}
	return list;
}
/**
 * @brief register all key binding categories to the appropriate widgets
 *
 * @return
 **/
void e2_keybinding_register_all (void)
{
	e2_keybinding_register (_C(16), app.main_window);  //_("general"

	gchar *category = g_strconcat(_C(16),".",_C(32),NULL);  //_("general.panes
	e2_keybinding_register (category, app.pane1_view.treeview);
	e2_keybinding_register (category, app.pane2_view.treeview);
	g_free (category);

	gchar *category2 = g_strconcat (_C(16),".",_C(4),NULL);  //_(general.command-line"
	gchar *category3 = g_strconcat (_C(16),".",_C(10),NULL);  //_(general.dir-line" 
	E2_CommandLineRuntime *rt;
	GList *line;
	for (line = cfg.command_lines; line != NULL ; line = g_list_next (line))
	{
		rt = (E2_CommandLineRuntime *) line->data;
		category = (rt->original) ? category2 : category3 ;
		e2_keybinding_register (category, GTK_BIN (rt->cl)->child);
	}
	g_free (category2);
	g_free (category3);
}
/**
 * @brief add key-binding data for @a category to categories list
 * Also initiates conversion of the actual bindings, when there is nothing
 * better to do, OR performs that translation, if E2_IDLE_KEYBINDING IS not defined
 * @param category internal name of the keybinding
 * @param widget the widget to which the keybinding will be attached
 * @return
 **/
void e2_keybinding_register (const gchar *category, GtkWidget *widget)
{
	//check if it already exists
	GList *list = e2_keybinding_find_by_name (category);
	if (list != NULL)
	{
		printd (DEBUG, "re-register keybinding category '%s'", category);
		E2_KeybindingRuntime *rt = list->data;
		(rt->ref_count)++;
		g_signal_connect (G_OBJECT (widget), "key-press-event",
			G_CALLBACK (_e2_keybinding_key_press_cb), rt);
		//remember what it's connected to, in case we need to disconnect later
		rt->instances = g_slist_append (rt->instances, widget);
		return;
	}
	//check if there is a parent category
	gchar *_parent = strrchr (category, '.');	//always ascii '.', don't need g_utf8_strrchr()
	GtkTreeIter iter_parent;
	if (_parent != NULL)
	{
		gchar *parent = g_strndup (category, _parent - category);
		GList *plist = e2_keybinding_find_by_name (parent);
		E2_KeybindingRuntime *rt = plist->data;
		GtkTreePath *path = gtk_tree_row_reference_get_path (rt->ref);
		if (path != NULL)
		{
			gtk_tree_model_get_iter (opt_keys->ex.tree.model, &iter_parent, path);
		}
		g_free (parent);
		gtk_tree_path_free (path);
	} else
	{
		printd (DEBUG, "create keybindings for category %s", category);
		gtk_tree_model_get_iter_first (opt_keys->ex.tree.model, &iter_parent);
	}
	printd (DEBUG, "register keybinding for category '%s'", category);
	const gchar *name = (((_parent != NULL) && (_parent[0] != '\0')) ?
		_parent + 1 : category);
	GtkTreeIter iter = iter_parent;
	if (!_e2_tree_find_iter_from_str (opt_keys->ex.tree.model, 0, name, &iter, FALSE))
	{
		gpointer options[1] = { (gpointer) name };
		_e2_option_tree_add (opt_keys, &iter, _parent != NULL ? &iter_parent : NULL,
			FALSE, TRUE, 1, options);
	}
	GtkTreePath *path = gtk_tree_model_get_path (opt_keys->ex.tree.model, &iter);

	//init runtime object
	E2_KeybindingRuntime *rt = g_malloc0 (sizeof (E2_KeybindingRuntime));
	rt->name = g_strdup (category);
	rt->model = opt_keys->ex.tree.model;
	rt->ref = gtk_tree_row_reference_new (opt_keys->ex.tree.model, path);
//	rt->keys = NULL;
	rt->ref_count = 2;	//CHECKME why needed?
//	rt->chained = NULL;
//	rt->timeout_id = 0;
//	g_object_set_data_full (G_OBJECT (widget), "e2-keybinding-runtime",
//		rt, (GDestroyNotify) widget_destroyed_cb);
	g_signal_connect (G_OBJECT (widget), "key-press-event",
		G_CALLBACK (_e2_keybinding_key_press_cb), rt);
	//remember what it's connected to, in case we need to disconnect later
	rt->instances = g_slist_append (rt->instances, widget);
	keys = g_list_append (keys, rt);
#ifdef E2_IDLE_KEYBINDING
	g_signal_connect (G_OBJECT (rt->model), "row-changed",
		G_CALLBACK (_e2_keybinding_model_changed_cb), NULL);
	//setup to convert this binding's data when there is nothing better to do
	g_idle_add ((GSourceFunc) _e2_keybinding_idle_sync_cb, NULL);
#else
	//translate the option tree data to rt form, now
	_e2_keybinding_sync (rt);
#endif
	//cleanup 
	gtk_tree_path_free (path);
}
/**
 * @brief cleanup key list and zap keys_cur ptr
 * @return
 **/
void e2_keybinding_clean (void)
{
	GList *categories;
	GSList *widgets;
	for (categories = keys; categories != NULL; categories = g_list_next (categories))
	{
		E2_KeybindingRuntime *rt = categories->data;
		g_free (rt->name);
		gtk_tree_row_reference_free (rt->ref);
		_e2_keybinding_free_keys (rt);
		//must disconnect signals that have data = rt, before freeing rt
		for (widgets = rt->instances; widgets != NULL; widgets = g_slist_next (widgets))
		{
			gulong handler = g_signal_handler_find (widgets->data, 
				G_SIGNAL_MATCH_FUNC, 0, 0, NULL, _e2_keybinding_key_press_cb, NULL);
			if (handler > 0)
				g_signal_handler_disconnect (widgets->data, handler);
		}
		g_slist_free (rt->instances);
		//CHECKME other cleans for contents of rt ?
		g_free (rt);
	}
	g_list_free (keys);
#ifdef E2_IDLE_KEYBINDING
	keys_cur =
#endif
	keys = NULL;
}
/**
 * @brief display current key bindings, all or a nominated section
 * @param parameters argument(s) to help command, starting with _(keys)
 * @return
 **/
void e2_keybinding_output_help (gchar *parameters)
{
	//decide which section of the bindings to report
	//NULL reports all (general)
	gchar *section = e2_utils_find_whitepace (parameters);
	if (section != NULL)
	{
		section =  g_strstrip (section);
		if (*section == '\0')
			section = NULL;
	}
	gchar *msg;
	// get treestore iter for start of bindings to report
	GtkTreeIter iter;
//	E2_OptionSet *opt_keys = e2_option_get ("keybindings");
	GtkTreeModel *mdl = opt_keys->ex.tree.model;
	if (section == NULL)
		gtk_tree_model_get_iter_first (mdl, &iter);
	else if (!e2_tree_find_iter_from_str (mdl, 0, section, &iter, FALSE))
	{
		msg = g_strdup_printf (_("Cannot find a key binding named %s"), section);
		e2_output_print_error(msg, TRUE);
		return;
		//revert to showing all
//		gtk_tree_model_get_iter_first (mdl, &iter);
	}

	msg = g_strdup_printf (
		"\t  "PROGNAME" %s\n"
		"\t+------------------------+\n",
		_("key bindings"));
	e2_output_print_default(msg);
	g_free (msg);
	
	gchar *cat, *key, *act, *arg;
	GtkTreeIter iter2, iter3;
	// get data from treestore
	//col 0 = category
	//col 1 = key
	//col 3 = action
	//col 4 = argument
	GString *str = g_string_new ("");
	do
	{
		if (gtk_tree_model_iter_children (mdl, &iter2, &iter))
		{
			do
			{
				gtk_tree_model_get (mdl, &iter2, 0, &cat, 1, &key, 3, &act, 4, &arg, -1);
				if (*cat != '\0')
				{
					g_string_append_printf (str, "\t%s\n", cat);
					g_free (cat);
					if (gtk_tree_model_iter_children (mdl, &iter3, &iter2))
					{
						do
						{
							gtk_tree_model_get (mdl, &iter3, 1, &key, 3, &act, 4, &arg, -1);
							g_string_append_printf (str, "\t\t%s\t%s\t%s\n", key, act, arg);
							g_free (key);
							g_free (act);
							g_free (arg);
						} while (gtk_tree_model_iter_next (mdl, &iter3));
					}
				}
				else
				{
					g_string_append_printf (str, "\t%s\t%s\t%s\n", key, act, arg);
					g_free (cat);
					g_free (key);
					g_free (act);
					g_free (arg);
				}
			} while (gtk_tree_model_iter_next (mdl, &iter2));
		}
	} while (gtk_tree_model_iter_next (mdl, &iter) && section == NULL);

	e2_output_print_default (str->str);
	g_string_free (str, TRUE);

#ifdef E2_VIKEYS
	if (e2_option_bool_get ("use-vi-keys"))
	{
		e2_output_print_default (
		_(	"Vi Keys\n~~~~~~~\n"
			"Updir\th\tOpen\tl\n"
			"Up\t\tk\tTop\t\tg\n"
			"Down\tj\tBottom\tG\n\n")
		);
	}
#endif
}
/**
 * @brief install default tree options for keybindings
 * This function is called only if the default is missing from the config file
 * @param set pointer to set data
 * @return
 **/
static void _e2_keybinding_tree_defaults (E2_OptionSet *set)
{
	//the key name strings are parsed by gtk, and except for simple letters
	//(which might reasonably change for different languages) no translation
	//is possible
	e2_option_tree_setup_defaults (set,
	g_strdup("<keybindings"),  //internal name
	g_strconcat(_C(16),"||||",NULL),  //_("general"
	g_strconcat("\t|F1||",_A(12),".",_A(63),"|",NULL), //"refresh"
	g_strconcat("\t|F2||",_A(5),".",_A(66),"|",NULL), //"rename"
	g_strconcat("\t|F3||",_A(5),".",_A(83),"|",NULL), //"view"
	g_strconcat("\t|<Shft>F3||",_A(5),".",_A(84),"|",NULL), //"view_again"
	g_strconcat("\t|<Ctrl>F3||",_A(5),".",_A(6),"|",NULL), //"find" (was F4)
	g_strconcat("\t|F4||",_A(5),".",_A(34),"|",NULL), //"edit"
	g_strconcat("\t|<Shft>F4||",_A(5),".",_A(35),"|",NULL), //"edit_again"
	g_strconcat("\t|F5||",_A(5),".",_A(30),"|",NULL),  //"copy"
	g_strconcat("\t|<Shft>F5||",_A(5),".",_A(31),"|",NULL),  //"copy_as"
	g_strconcat("\t|F6||",_A(5),".",_A(52),"|",NULL), //"move"
	g_strconcat("\t|<Shft>F6||",_A(5),".",_A(53),"|",NULL), //"move_as"
	g_strconcat("\t|F7||",_A(3),".",_A(50),"|",NULL), //"mkdir"
	g_strconcat("\t|F8||",_A(5),".",_A(81),"|",NULL), //"trash"
	g_strconcat("\t|<Ctrl>F8||",_A(6),".",_A(81),"|",NULL), //"goto trash"  modifiers not translated, gtk needs english to parse
	g_strconcat("\t|<Shft><Ctrl>F8||",_A(5),".",_A(33),"|",NULL), //"delete"
	g_strconcat("\t|F9||",_A(5),".",_A(45),"|",NULL), //"file info"
	g_strconcat("\t|F10||",_A(1),".",_A(62),"|",NULL), //"quit"
	g_strconcat("\t|F11||",_A(2),".",_A(29),"|",NULL),  //"configure"
	g_strconcat("\t|F12||",_A(12),".",_A(76),"|",NULL), //"toggle direction"
#ifdef E2_OUTPUTTABS
	g_strconcat("\t|<Ctrl>F12||",_A(8),".",_A(24),"|",NULL), //"output.add
	g_strconcat("\t|<Shft><Ctrl>F12||",_A(8),".",_A(33),"|",NULL), //"output.delete
#endif
	//conform this help string to the following 'jump-keys'
	g_strconcat("\t|<Ctrl>",_("g"),"||",_A(8),".",_A(61),"|",_("Now press one of h,m,d\\n"),NULL), //"print"
	//these 3 children are the so-called "directory jump keys"
	g_strconcat("\t\t|",_("h"),"||cd|$HOME",NULL), //_A(15)
	g_strconcat("\t\t|",_("m"),"||cd|/mnt",NULL),  //_A(15)
	g_strconcat("\t\t|",_("d"),"||cd|$HOME/downloads",NULL),  //_A(15)
	g_strconcat("\t|<Ctrl>",_("w"),"||",_A(8),".",_A(25),"|*,1",NULL),  //"adjust ratio" no arg translation
	g_strconcat("\t|<Ctrl>",_("e"),"||",_A(8),".",_A(25),"|*,0",NULL),    //"adjust ratio" no arg translation
	g_strconcat("\t|<Ctrl>",_("z"),"||",_A(1),".",_A(38),"|",NULL), //"focus toggle"
	g_strconcat("\t|<Ctrl>",_("r"),"||",_A(12),".",_A(63),"|",NULL), //"refresh"
	g_strconcat("\t|<Ctrl>","1","||",_A(4),".",_A(38),"|1",NULL), //"focus toggle" no arg translation
	g_strconcat("\t|<Ctrl>","2","||",_A(4),".",_A(38),"|2",NULL), //"focus toggle"no arg translation
	g_strconcat("\t",_C(32),"||||",NULL),  //_("panes"
	//note "Space" not recognised, must be "space"
//	g_strconcat("\t\t|space||",_A(9),".",_A(72),"|",NULL), //"switch"  something else makes this 'open' the selected item
	g_strconcat("\t\t|space||",_A(9),".",_A(80),"|",NULL), //"toggle selected"
	g_strconcat("\t\t|Tab||",_A(9),".",_A(72),"|",NULL), //"switch"
	g_strconcat("\t\t|Return||",_A(5),".",_A(54),"|",NULL), //"open"
	g_strconcat("\t\t|<Shft>Return||",_A(5),".",_A(55),"|",NULL), //"open in other"
	g_strconcat("\t\t|Left||",_A(9),".",_A(39),"|",NULL), //"go back"
	g_strconcat("\t\t|Right||",_A(9),".",_A(40),"|",NULL), //"go forward"
	g_strconcat("\t\t|<Alt>Up||cd|..",NULL),  //_A(15)
	g_strconcat("\t\t|BackSpace||cd|..",NULL),  //_A(15)
	g_strconcat("\t\t|Delete||",_A(5),".",_A(33),"|",NULL), //"delete"
	g_strconcat("\t\t|<Shft>Delete||",_A(1),".",_A(82),"|",NULL), //"trashempty"
	g_strconcat("\t\t|Home||cd|$HOME",NULL),   //_A(15)
	g_strconcat("\t\t|<Ctrl>",_("a"),"||",_A(9),".",_A(79),"|",NULL), //"toggle select all"
	g_strconcat("\t\t|<Ctrl>",_("i"),"||",_A(9),".",_A(47),"|",NULL), //"invert selection"
	g_strconcat("\t\t|<Ctrl>",_("f"),"||",_A(5),".",_A(6),"|",NULL), //"find"
	g_strconcat("\t\t|<Ctrl>",_("x"),"||",_A(12),".",_A(75),"|",NULL), //"sync"
	g_strconcat("\t\t|<Ctrl>",_("h"),"||",_A(9),".",_A(78),"|",NULL),  //"toggle hidden"
//	g_strconcat("\t\t|/||",_A(5),".",_A(6),"|",NULL),  //"find"
	g_strconcat("\t\t|<Ctrl>Left||",_A(12),".",_A(25),"|","*,1",NULL), //"adjust ratio" no arg translation
	g_strconcat("\t\t|<Ctrl>Right||",_A(12),".",_A(25),"|","*,0",NULL), //"adjust ratio" no arg translation
	g_strconcat("\t\t|<Ctrl>",_("p"),"||",_A(9),".",_A(71),"|",NULL),  //"show menu"
	g_strconcat("\t\t|<Shft><Ctrl>",_("p"),"||",_A(9),".",_A(71),"|",_A(93),NULL), //"show menu, shift"
	g_strconcat("\t\t|<Ctrl><Alt>",_("p"),"||",_A(9),".",_A(71),"|",_A(87),NULL),  //"show menu, ctrl"
	g_strconcat("\t",_C(10),"||||",NULL),  //_"directory line"
	g_strconcat("\t\t|Tab||",_A(1),".",_A(28),"|",_A(88),NULL), //"complete dirs"
	g_strconcat("\t\t|<Ctrl>Tab||",_A(12),".",_A(54),"|",NULL), //"open"
	g_strconcat("\t\t|<Ctrl>Delete||",_A(1),".",_A(26),"|",NULL),  //"clear"
	g_strconcat("\t\t|<Alt>Delete||",_A(1),".",_A(27),"|",NULL),  //"clear_history"
	g_strconcat("\t\t|<Ctrl>Return||",_A(1),".",_A(46),"|",_A(89),NULL), //"insert selection, escape"
	g_strconcat("\t\t|<Alt>Return||",_A(1),".",_A(46),"|",NULL), //"insert selection"
	g_strconcat("\t",_C(4),"||||",NULL),  //_("command line"
	g_strconcat("\t\t|<Ctrl>Return||",_A(1),".",_A(46),"|",_A(89),NULL), //"insert selection, escape"
	g_strconcat("\t\t|<Alt>Return||",_A(1),".",_A(46),"|",_A(92),NULL), //"insert selection, quote"
	g_strconcat("\t\t|<Ctrl>Delete||",_A(1),".",_A(26),"|",NULL), //"clear"
	g_strconcat("\t\t|<Alt>Delete||",_A(1),".",_A(27),"|",NULL),  //"clear_history"
	g_strconcat("\t\t|Tab||",_A(1),".",_A(28),"|",NULL), //"complete"
//	g_strconcat("\t",_C(27),"||||",NULL),  //_("output"
//CHECKME these stay as command-line bindings ??
	g_strconcat("\t\t|Page_Up||",_A(8),".",_A(59),"|",NULL), //"page up"
	g_strconcat("\t\t|Page_Down||",_A(8),".",_A(58),"|",NULL), //"page down"
	g_strconcat("\t\t|<Shft>Page_Up||",_A(8),".",_A(43),"|",NULL), //"goto top"
	g_strconcat("\t\t|<Shft>Page_Down||",_A(8),".",_A(42),"|",NULL), //"goto bottom"
	g_strconcat("\t\t|<Shft>Up||",_A(8),".",_A(68),"|",NULL),  //"scroll up"
	g_strconcat("\t\t|<Shft>Down||",_A(8),".",_A(67),"|",NULL),  //"scroll down"
	g_strdup(">"),
	NULL);
}
/**
 * @brief initialize key-related options, and init keys pointer
 * @return
 **/
void e2_keybinding_options_register (void)
{
	//no screen rebuilds needed after any change to these options
	gchar *group_name = g_strconcat(_C(19),".",_C(21),NULL);
	opt_keys = e2_option_tree_register ("keybindings", group_name, _C(21),  //_("keybindings"
		NULL, _e2_keybinding_tree_selection_check_cb, _e2_keybinding_tree_draggable_check_cb,
		E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP | E2_OPTION_FLAG_BUILDKEYS);
	e2_option_tree_add_column (opt_keys, _("Category"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, E2_OPTION_TREE_COL_NOT_EDITABLE, NULL, NULL);
	e2_option_tree_add_column (opt_keys, _("Key"), E2_OPTION_TREE_TYPE_KEY, 0, "",
		NULL, NULL, 0, _e2_keybinding_visible_check_cb, NULL);
	e2_option_tree_add_column (opt_keys, _("Continue"), E2_OPTION_TREE_TYPE_BOOL, FALSE, "false", //no translation
		NULL, NULL, 0, _e2_keybinding_visible_check_cb, NULL);
	e2_option_tree_add_column (opt_keys, _("Action"), E2_OPTION_TREE_TYPE_SEL, 0, NULL,
		e2_actions_fill_store, GINT_TO_POINTER (E2_ACTION_EXCLUDE_ACCEL), 0,
		_e2_keybinding_visible_check_cb, NULL);
	e2_option_tree_add_column (opt_keys, _("Argument"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, _e2_keybinding_visible_check_cb, NULL);
	e2_option_tree_sync_columns (opt_keys);

	e2_option_tree_prepare_defaults (opt_keys, _e2_keybinding_tree_defaults);

	group_name = g_strconcat(_C(19),":",_C(24),NULL); //_("interface:miscellaneous"
	e2_option_int_register ("keybindings-timeout", group_name, _("chained keybindings timeout (ms)"), 
		_("This sets the time limit (in milliseconds) for accepting 'chained' keybindings"),
		NULL, 2000, 1, 1000000,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
}
